import logging

from aws_cloudformation_power_switch.power_switch import PowerSwitch
from aws_cloudformation_power_switch.tag import logical_id, stack_name


class EC2PowerSwitch(PowerSwitch):
    def __init__(self):
        super(EC2PowerSwitch, self).__init__()

    def startup(self, instance: dict):
        instance_id = instance["InstanceId"]
        logging.info("starting ec2 instance %s (%s)", logical_id(instance), instance_id)
        if not self.dry_run:
            self.ec2.start_instances(InstanceIds=[instance_id])

    def instance_state(self, instance) -> str:
        return instance.get("State", {}).get("Name", "unknown")

    def instance_id(self, instance) -> str:
        return instance.get("InstanceId")

    def instance_needs_shutdown(self, instance) -> bool:
        return self.instance_state(instance) == "running"

    def instance_needs_startup(self, instance) -> bool:
        return self.instance_state(instance) == "stopped"

    def shutdown(self, instance: dict):
        instance_id = instance["InstanceId"]
        logging.info(
            "shutting down ec2 instance %s (%s)", logical_id(instance), instance_id
        )
        if not self.dry_run:
            self.ec2.stop_instances(InstanceIds=[instance_id])

    @property
    def ec2(self):
        return self.session.client("ec2")

    def select_instances(self):
        result = []
        paginator = self.ec2.get_paginator("describe_instances")
        for response in paginator.paginate():
            for reservation in response["Reservations"]:
                for instance in reservation["Instances"]:
                    if stack_name(instance).startswith(self.stack_name_prefix):
                        result.append(instance)

        result = sorted(result, key=lambda instance: logical_id(instance))

        for i in filter(lambda i: self.verbose, result):
            logging.info(
                "found ec2 instance %s (%s) in state %s",
                logical_id(i),
                i["InstanceId"],
                i.get("State", {}).get("Name", "unknown"),
            )

        if not result and self.verbose:
            logging.info("no EC2 instances found")

        return result



import os
import jsonschema
import logging
from .master import power_switch

schema = {
    "type": "object",
    "required": ["stack_name_prefix"],
    "properties": {
        "stack_name_prefix": {"type": "string"},
        "profile": {"type": "string"},
        "region": {"type": "string"},
        "verbose": {"type": "boolean", "default": True},
        "dry_run": {"type": "boolean", "default": False},
        "state": {"type": "string", "enum": ["on", "off"]},
    },
}


def is_valid_request(request) -> bool:
    try:
        jsonschema.validate(request, schema)
        return True
    except jsonschema.ValidationError as e:
        logging.error("invalid request received: %s" % str(e.context))
        return False


def handler(request, context):
    logging.getLogger().setLevel(os.getenv("LOG_LEVEL", "INFO"))
    if is_valid_request(request):
        state = request["state"]
        request.pop("state")
        switch = power_switch(**request)
        if state == "off":
            switch.off()
        else:
            switch.on()



import logging

from botocore.exceptions import ClientError

from aws_cloudformation_power_switch.power_switch import PowerSwitch
from aws_cloudformation_power_switch.tag import logical_id


class RDSClusterPowerSwitch(PowerSwitch):
    def __init__(self):
        super(RDSClusterPowerSwitch, self).__init__()
        self.resource_type = "AWS::RDS::DBCluster"

    def startup(self, instance: dict):
        name = logical_id(instance)
        cluster_id = self.instance_id(instance)
        logging.info("startup rds cluster %s", cluster_id)
        if not self.dry_run:
            try:
                self.rds.start_db_cluster(DBClusterIdentifier=cluster_id)
            except ClientError as e:
                logging.error("failed to stop %s (%s), %s", name, cluster_id, e)

    def shutdown(self, instance: dict):
        name = logical_id(instance)
        cluster_id = self.instance_id(instance)
        logging.info("shutdown rds cluster %s (%s)", name, cluster_id)
        if not self.dry_run:
            try:
                self.rds.stop_db_cluster(DBClusterIdentifier=cluster_id)
            except ClientError as e:
                logging.error("failed to stop %s (%s), %s", name, cluster_id, e)

    def instance_id(self, instance) -> str:
        return instance["DBClusterIdentifier"]

    def instance_state(self, instance) -> str:
        return instance["Status"]

    def instance_needs_shutdown(self, instance) -> bool:
        return self.instance_state(instance) == "available"

    def instance_needs_startup(self, instance) -> bool:
        return self.instance_state(instance) == "stopped"

    @property
    def rds(self):
        return self.session.client("rds")

    def select_instances(self):
        result = []
        if self.rds.describe_db_clusters().get("DBClusters"):
            for r in self.stack_resources:
                instance = self.rds.describe_db_clusters(
                    DBClusterIdentifier=r["PhysicalResourceId"]
                )["DBClusters"][0]
                instance["TagList"] = [
                    {"Key": "aws:cloudformation:stack-name", "Value": r["StackName"]},
                    {
                        "Key": "aws:cloudformation:logical-id",
                        "Value": r["LogicalResourceId"],
                    },
                ]
                result.append(instance)

        for i in filter(lambda i: self.verbose, result):
            logging.info(
                "rds cluster %s (%s) in state %s",
                logical_id(i),
                i["DBClusterIdentifier"],
                i["Status"],
            )

        if not result and self.verbose:
            logging.info("No RDS clusters found")

        return result



import os
import logging
import click

from aws_cloudformation_power_switch.master import power_switch


@click.group(help="power off all running instances managed by CloudFormation stacks")
@click.pass_context
@click.option(
    "--stack-name-prefix", "-s", required=True, help="of stacks to startup/shutdown"
)
@click.option(
    "--dry-run",
    "-d",
    is_flag=True,
    default=False,
    help="do not change anything, just show what is going to happen",
)
@click.option("--profile", "-p", help="to use")
@click.option("--region", "-r", help="to use")
@click.option("--verbose", "-v", is_flag=True, default=False, help="output")
def main(ctx, **kwargs):
    logging.basicConfig(level=os.getenv("LOG_LEVEL", "INFO"))
    ctx.obj = power_switch(**kwargs)


@main.command(name="on", help="start all ec2, ecs, rds and auto scaling instances")
@click.pass_context
def on(ctx):
    ctx.obj.on()


@main.command(name="off", help="stop all ec2, ecs, rds and auto scaling instances")
@click.pass_context
def off(ctx):
    ctx.obj.off()


if __name__ == "__main__":
    main()



import logging

from botocore.exceptions import ClientError

from aws_cloudformation_power_switch.power_switch import PowerSwitch
from aws_cloudformation_power_switch.tag import logical_id, stack_name


class RDSPowerSwitch(PowerSwitch):
    def __init__(self):
        super(RDSPowerSwitch, self).__init__()

    def startup(self, instance: dict):
        db = instance["DBInstanceIdentifier"]
        logging.info("startup rds instance %s", db)
        if not self.dry_run:
            try:
                self.rds.start_db_instance(DBInstanceIdentifier=db)
            except ClientError as e:
                logging.error("failed to start %s, %s", db, e)

    def shutdown(self, instance: dict):
        db = instance["DBInstanceIdentifier"]
        name = logical_id(instance)
        logging.info("shutdown rds instance %s (%s)", name, db)
        if not self.dry_run:
            try:
                self.rds.stop_db_instance(DBInstanceIdentifier=db)
            except ClientError as e:
                logging.error("failed to stop %s, %s", db, e)

    def instance_state(self, instance) -> str:
        return instance["DBInstanceStatus"]

    def instance_needs_shutdown(self, instance) -> bool:
        return self.instance_state(instance) == "available"

    def instance_needs_startup(self, instance) -> bool:
        return self.instance_state(instance) == "stopped"

    @property
    def rds(self):
        return self.session.client("rds")

    def select_instances(self):
        result = []
        paginator = self.rds.get_paginator("describe_db_instances")
        for response in paginator.paginate():
            for instance in response["DBInstances"]:
                arn = instance["DBInstanceArn"]
                instance.update(self.rds.list_tags_for_resource(ResourceName=arn))
                if "DBClusterIdentifier" not in instance and stack_name(
                    instance
                ).startswith(self.stack_name_prefix):
                    result.append(instance)

        result = sorted(result, key=lambda instance: logical_id(instance))

        for i in filter(lambda i: self.verbose, result):
            logging.info(
                "rds instance %s (%s) in state %s"
                % (logical_id(i), i["DBInstanceIdentifier"], i["DBInstanceStatus"])
            )

        if not result and self.verbose:
            logging.info("No RDS instances found")

        return result



def value(instance: dict, key: str) -> str:
    """
    gets the first value of the AWS object `instance` tag, or "" if not found
    """
    tags = instance.get("Tags", instance.get("TagList", []))
    return next(
        iter(map(lambda c: c["Value"], filter(lambda t: t["Key"] == key, tags))), ""
    )


def stack_name(instance: dict) -> str:
    return value(instance, "aws:cloudformation:stack-name")


def logical_id(instance: dict) -> str:
    return value(instance, "aws:cloudformation:logical-id")



from .asg import ASGPowerSwitch
from .ec2 import EC2PowerSwitch
from .power_switch import PowerSwitch
from .rds import RDSPowerSwitch
from .rds_cluster import RDSClusterPowerSwitch
from .ecs import ECSPowerSwitch


class MasterPowerSwitch(PowerSwitch):
    def __init__(self):
        super(MasterPowerSwitch, self).__init__()

    def select_instances(self):
        result = [
            ASGPowerSwitch(),
            EC2PowerSwitch(),
            RDSClusterPowerSwitch(),
            RDSPowerSwitch(),
            ECSPowerSwitch(),
        ]
        for switch in result:
            switch.session = self.session
            switch.dry_run = self.dry_run
            switch.stack_name_prefix = self.stack_name_prefix
            switch.verbose = self.verbose
        return result

    def instance_needs_startup(self, instance) -> bool:
        return True

    def instance_needs_shutdown(self, instance) -> bool:
        return True

    def startup(self, instance: PowerSwitch):
        instance.on()

    def shutdown(self, instance: PowerSwitch):
        instance.off()


def power_switch(
    stack_name_prefix, dry_run=True, verbose=False, profile=None, region=None
):
    result = MasterPowerSwitch()
    result.set_session(profile, region)
    result.dry_run = dry_run
    result.verbose = verbose
    result.stack_name_prefix = stack_name_prefix
    assert stack_name_prefix, "stack name prefix is required"
    return result



import boto3
import sys
from typing import List, NamedTuple
import logging
import sys


class PowerSwitch(object):
    def __init__(self):
        self.stack_name_prefix = None
        self.verbose = False
        self.dry_run = True
        self.session = None
        self._stack_resources = []
        self.resource_type = None

    def select_instances(self) -> List:
        if not self.resource_type:
            raise Exception("no CloudFormation resource type specified")
        return self.stack_resources

    def instance_needs_startup(self, instance) -> bool:
        raise Exception("not implemented")

    def instance_needs_shutdown(self, instance) -> bool:
        raise Exception("not implemented")

    def shutdown(self, instance):
        raise Exception("not implemented")

    def startup(self, instance):
        raise Exception("not implemented")

    def check_precondition(self):
        if not self.stack_name_prefix:
            logging.error("no stack name prefix specified")
            sys.exit(1)

    def on(self):
        self.check_precondition()
        for instance in filter(
            lambda i: self.instance_needs_startup(i), self.select_instances()
        ):
            self.startup(instance)

    def off(self):
        self.check_precondition()
        for instance in filter(
            lambda i: self.instance_needs_shutdown(i), self.select_instances()
        ):
            self.shutdown(instance)

    def set_session(self, profile, region):
        kwargs = {}
        if region:
            kwargs["region_name"] = region
        if profile:
            kwargs["profile_name"] = profile

        self.session = boto3.Session(**kwargs)

    def get_stack_resources(self):
        if self.session in stack_resources_per_session:
            self._stack_resources = stack_resources_per_session[self.session]
            return

        self._stack_resources = []
        logging.info(
            "retrieving all CloudFormation stacks with prefix %s",
            self.stack_name_prefix,
        )
        cfn = self.session.client("cloudformation")
        for response in cfn.get_paginator("describe_stacks").paginate():
            for stack_summary in response["Stacks"]:
                name = stack_summary["StackName"]
                if name.startswith(self.stack_name_prefix):
                    if self.verbose:
                        logging.info("loading all resources from stack %s", name)
                    for response in cfn.get_paginator("list_stack_resources").paginate(
                        StackName=name
                    ):
                        for r in response["StackResourceSummaries"]:
                            r["StackName"] = name
                        self._stack_resources.extend(response["StackResourceSummaries"])

        stack_resources_per_session[self.session] = self._stack_resources

    @property
    def stack_resources(self) -> List[dict]:
        if not self._stack_resources:
            self.get_stack_resources()

        return sorted(
            filter(
                lambda r: not self.resource_type
                or r.get("ResourceType") == self.resource_type,
                self._stack_resources,
            ),
            key=lambda r: r["StackName"] + "-" + r["LogicalResourceId"],
        )

    def get_stack_resource_by_physical_id(self, physical_resource_id: str):
        next(
            filter(
                lambda r: r["PhysicalResourceId"] == physical_resource_id,
                self.stack_resources,
            ),
            None,
        )


stack_resources_per_session = {}



import logging
from typing import List


from aws_cloudformation_power_switch.power_switch import PowerSwitch
from aws_cloudformation_power_switch.tag import stack_name, logical_id


class ASGPowerSwitch(PowerSwitch):
    def __init__(self):
        super(ASGPowerSwitch, self).__init__()

    def startup(self, instance: dict):
        name = instance["AutoScalingGroupName"]
        max_size = instance["MaxSize"]

        logging.info(
            "setting capacity of auto scaling group %s (%s) to %s",
            logical_id(instance),
            name,
            max_size,
        )
        if not self.dry_run:
            self.autoscaling.update_auto_scaling_group(
                AutoScalingGroupName=name, DesiredCapacity=max_size
            )

    def shutdown(self, instance: dict):
        name = instance["AutoScalingGroupName"]
        logging.info(
            "setting capacity of auto scaling group %s (%s) to 0",
            logical_id(instance),
            name,
        )
        if not self.dry_run:
            self.autoscaling.update_auto_scaling_group(
                AutoScalingGroupName=name, DesiredCapacity=0, MinSize=0
            )

    def instance_desired_capacity(self, instance) -> int:
        return instance["DesiredCapacity"]

    def instance_needs_shutdown(self, instance) -> bool:
        return self.instance_desired_capacity(instance) > 0

    def instance_needs_startup(self, instance) -> bool:
        return self.instance_desired_capacity(instance) == 0

    @property
    def autoscaling(self):
        return self.session.client("autoscaling")

    def select_instances(self) -> List[dict]:
        result = []
        paginator = self.autoscaling.get_paginator("describe_auto_scaling_groups")
        for response in paginator.paginate():
            result.extend(
                filter(
                    lambda a: stack_name(a).startswith(self.stack_name_prefix),
                    response["AutoScalingGroups"],
                )
            )

        for i in filter(lambda i: self.verbose, result):
            logging.info(
                "found auto scaling group %s (%s) with desired capacity %s",
                logical_id(i),
                i["AutoScalingGroupName"],
                i["DesiredCapacity"],
            )

        if not result and self.verbose:
            logging.info("No auto scaling groups found")

        return result



import logging
import re
from typing import Iterator

from aws_cloudformation_power_switch.power_switch import PowerSwitch
from aws_cloudformation_power_switch.tag import logical_id, stack_name


class ECSPowerSwitch(PowerSwitch):
    def __init__(self):
        super(ECSPowerSwitch, self).__init__()

    def startup(self, instance: dict):
        desired_count = self.find_last_steady_state_count(instance)
        logging.info(
            "starting ecs service %s with desired count = %d",
            self.instance_id(instance),
            desired_count,
        )
        if not self.dry_run:
            self.ecs.update_service(
                service=self.instance_id(instance),
                cluster=self.cluster_id(instance),
                desiredCount=desired_count,
            )

    def desired_count(self, instance) -> int:
        return instance.get("desiredCount")

    def instance_id(self, instance) -> str:
        return instance.get("serviceArn")

    def cluster_id(self, instance) -> str:
        return instance.get("clusterArn")

    def instance_needs_shutdown(self, instance) -> bool:
        return self.desired_count(instance) > 0

    def instance_needs_startup(self, instance) -> bool:
        return self.desired_count(instance) == 0

    def find_last_steady_state_count(self, instance: dict) -> int:
        pattern = re.compile(r"has\s+started\s+(?P<count>[1-9][0-9]*)\s+tasks")
        for event in instance["events"]:
            match = pattern.search(event["message"])
            if match:
                return int(match.group("count"))
        return 0

    def shutdown(self, instance: dict):
        logging.info("shutting down ecs service %s", self.instance_id(instance))
        if not self.dry_run:
            self.ecs.update_service(
                service=self.instance_id(instance),
                cluster=self.cluster_id(instance),
                desiredCount=0,
            )

    @property
    def ecs(self):
        return self.session.client("ecs")

    def _list_clusters(self) -> Iterator[str]:
        resourcegroups = self.session.client("resourcegroupstaggingapi")
        paginator = resourcegroups.get_paginator("get_resources")
        for resources in paginator.paginate(
            ResourceTypeFilters=["ecs:cluster"],
            TagFilters=[{"Key": "aws:cloudformation:stack-name"}],
        ):
            for resource in resources["ResourceTagMappingList"]:
                if stack_name(resource).startswith(self.stack_name_prefix):
                    yield resource["ResourceARN"]

    def select_instances(self):
        result = []
        for cluster_arn in self._list_clusters():
            paginator = self.ecs.get_paginator("list_services")
            for services in paginator.paginate(cluster=cluster_arn, maxResults=10):
                response = self.ecs.describe_services(
                    services=services["serviceArns"], cluster=cluster_arn
                )
                result.extend(response["services"])

        result = sorted(result, key=lambda instance: instance["serviceName"])

        for i in filter(lambda i: self.verbose, result):
            logging.info(
                "found ecs service %s (%s) with desired count set to %s",
                logical_id(i),
                i["serviceName"],
                i["desiredCount"],
            )

        if not result and self.verbose:
            logging.info("no ECS services found")

        return result
